@@ -6,6 +6,11 @@ from functools import wraps
 from velocity.db import exceptions
 from velocity.db.core.transaction import Transaction
 
+import logging
+
+logger = logging.getLogger("velocity.db.engine")
+logger.setLevel(logging.INFO)  # Or DEBUG for more verbosity
+
 
 class Engine:
     """
@@ -14,10 +19,11 @@ class Engine:
 
     MAX_RETRIES = 100
 
-    def __init__(self, driver, config, sql):
+    def __init__(self, driver, config, sql, connect_timeout=5):
         self.__config = config
         self.__sql = sql
         self.__driver = driver
+        self.__connect_timeout = connect_timeout
 
     def __str__(self):
         return f"[{self.sql.server}] engine({self.config})"
@@ -39,16 +45,40 @@ class Engine:
     def __connect(self):
         """
         Internal connection logic, raising suitable exceptions on error.
+        Enforces a connect timeout and handles different config types.
         """
+        server = self.sql.server.lower()
+        timeout_key = "timeout" if "sqlite" in server else "connect_timeout"
+        timeout_val = self.__connect_timeout
+
         try:
             if isinstance(self.config, dict):
-                return self.driver.connect(**self.config)
-            if isinstance(self.config, (tuple, list)):
-                return self.driver.connect(*self.config)
-            if isinstance(self.config, str):
-                return self.driver.connect(self.config)
-            raise Exception("Unhandled configuration parameter.")
-        except:
+                config = self.config.copy()
+                if timeout_key not in config:
+                    config[timeout_key] = timeout_val
+                return self.driver.connect(**config)
+
+            elif isinstance(self.config, str):
+                conn_str = self.config
+                if timeout_key not in conn_str:
+                    conn_str += f" {timeout_key}={timeout_val}"
+                return self.driver.connect(conn_str)
+
+            elif isinstance(self.config, (tuple, list)):
+                config_args = list(self.config)
+                if config_args and isinstance(config_args[-1], dict):
+                    if timeout_key not in config_args[-1]:
+                        config_args[-1][timeout_key] = timeout_val
+                else:
+                    config_args.append({timeout_key: timeout_val})
+                return self.driver.connect(*config_args)
+
+            else:
+                raise TypeError(
+                    f"Unhandled configuration parameter type: {type(self.config)}"
+                )
+
+        except Exception:
             self.process_error()
 
     def transaction(self, func_or_cls=None):
@@ -57,6 +87,8 @@ class Engine:
         May also be used to decorate a class, in which case all methods are wrapped in a transaction if they accept `tx`.
         With no arguments, returns a new Transaction directly.
         """
+        # print("Transaction", func_or_cls.__name__, type(func_or_cls))
+
         if func_or_cls is None:
             return Transaction(self)
 
@@ -64,15 +96,21 @@ class Engine:
             return classmethod(self.transaction(func_or_cls.__func__))
 
         if inspect.isfunction(func_or_cls) or inspect.ismethod(func_or_cls):
+            names = list(inspect.signature(func_or_cls).parameters.keys())
+            # print(func_or_cls.__name__, names)
+            if "_tx" in names:
+                raise NameError(
+                    f"In function {func_or_cls.__name__}, '_tx' is not allowed as a parameter."
+                )
 
             @wraps(func_or_cls)
             def new_function(*args, **kwds):
                 tx = None
                 names = list(inspect.signature(func_or_cls).parameters.keys())
-                if "_tx" in names:
-                    raise NameError(
-                        f"In function {func_or_cls.__name__}, '_tx' is not allowed as a parameter."
-                    )
+
+                # print("inside", func_or_cls.__name__)
+                # print(names)
+                # print(args, kwds)
 
                 if "tx" not in names:
                     # The function doesn't even declare a `tx` parameter, so run normally.
@@ -91,33 +129,31 @@ class Engine:
                     if len(args) > pos:
                         if isinstance(args[pos], Transaction):
                             tx = args[pos]
-                        else:
-                            raise TypeError(
-                                f"In function {func_or_cls.__name__}, positional argument `tx` must be a Transaction object."
-                            )
+
                 if tx:
                     return self.exec_function(func_or_cls, tx, *args, **kwds)
+
                 with Transaction(self) as local_tx:
                     pos = names.index("tx")
-                    args = list(args)
-                    args.insert(pos, local_tx)
-                    args = tuple(args)
-                    return self.exec_function(func_or_cls, local_tx, *args, **kwds)
+                    new_args = args[:pos] + (local_tx,) + args[pos:]
+                    return self.exec_function(func_or_cls, local_tx, *new_args, **kwds)
 
             return new_function
 
         if inspect.isclass(func_or_cls):
 
-            class NewCls(func_or_cls):
-                pass
+            NewCls = type(func_or_cls.__name__, (func_or_cls,), {})
+
+            for attr_name in dir(func_or_cls):
+                # Optionally skip special methods
+                if attr_name.startswith("__") and attr_name.endswith("__"):
+                    continue
+
+                attr = getattr(func_or_cls, attr_name)
+
+                if callable(attr):
+                    setattr(NewCls, attr_name, self.transaction(attr))
 
-            for attr_name, attr_value in func_or_cls.__dict__.items():
-                if (
-                    isinstance(attr_value, classmethod)
-                    or inspect.isfunction(attr_value)
-                    or inspect.ismethod(attr_value)
-                ):
-                    setattr(NewCls, attr_name, self.transaction(attr_value))
             return NewCls
 
         return Transaction(self)
@@ -301,67 +337,69 @@ class Engine:
 
         error_code, error_mesg = self.sql.get_error(e)
 
+        logger.warning(
+            "Database error caught. Attempting to transform: code=%s message=%s",
+            error_code,
+            error_mesg,
+        )
+
         if error_code in self.sql.ApplicationErrorCodes:
-            raise exceptions.DbApplicationError(e)
+            raise exceptions.DbApplicationError from None
         if error_code in self.sql.ColumnMissingErrorCodes:
-            raise exceptions.DbColumnMissingError(e)
+            raise exceptions.DbColumnMissingError from None
         if error_code in self.sql.TableMissingErrorCodes:
-            raise exceptions.DbTableMissingError(e)
+            raise exceptions.DbTableMissingError from None
         if error_code in self.sql.DatabaseMissingErrorCodes:
-            raise exceptions.DbDatabaseMissingError(e)
+            raise exceptions.DbDatabaseMissingError from None
         if error_code in self.sql.ForeignKeyMissingErrorCodes:
-            raise exceptions.DbForeignKeyMissingError(e)
+            raise exceptions.DbForeignKeyMissingError from None
         if error_code in self.sql.TruncationErrorCodes:
-            raise exceptions.DbTruncationError(e)
+            raise exceptions.DbTruncationError from None
         if error_code in self.sql.DataIntegrityErrorCodes:
-            raise exceptions.DbDataIntegrityError(e)
+            raise exceptions.DbDataIntegrityError from None
         if error_code in self.sql.ConnectionErrorCodes:
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if error_code in self.sql.DuplicateKeyErrorCodes:
-            raise exceptions.DbDuplicateKeyError(e)
+            raise exceptions.DbDuplicateKeyError from None
         if re.search(r"key \(sys_id\)=\(\d+\) already exists.", msg, re.M):
-            raise exceptions.DbDuplicateKeyError(e)
+            raise exceptions.DbDuplicateKeyError from None
         if error_code in self.sql.DatabaseObjectExistsErrorCodes:
-            raise exceptions.DbObjectExistsError(e)
+            raise exceptions.DbObjectExistsError from None
         if error_code in self.sql.LockTimeoutErrorCodes:
-            raise exceptions.DbLockTimeoutError(e)
+            raise exceptions.DbLockTimeoutError from None
         if error_code in self.sql.RetryTransactionCodes:
-            raise exceptions.DbRetryTransaction(e)
+            raise exceptions.DbRetryTransaction from None
         if re.findall(r"database.*does not exist", msg, re.M):
-            raise exceptions.DbDatabaseMissingError(e)
+            raise exceptions.DbDatabaseMissingError from None
         if re.findall(r"no such database", msg, re.M):
-            raise exceptions.DbDatabaseMissingError(e)
+            raise exceptions.DbDatabaseMissingError from None
         if re.findall(r"already exists", msg, re.M):
-            raise exceptions.DbObjectExistsError(e)
+            raise exceptions.DbObjectExistsError from None
         if re.findall(r"server closed the connection unexpectedly", msg, re.M):
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if re.findall(r"no connection to the server", msg, re.M):
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if re.findall(r"connection timed out", msg, re.M):
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if re.findall(r"could not connect to server", msg, re.M):
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if re.findall(r"cannot connect to server", msg, re.M):
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if re.findall(r"connection already closed", msg, re.M):
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if re.findall(r"cursor already closed", msg, re.M):
-            raise exceptions.DbConnectionError(e)
+            raise exceptions.DbConnectionError from None
         if "no such table:" in msg:
-            raise exceptions.DbTableMissingError(e)
+            raise exceptions.DbTableMissingError from None
 
-        msg = f"""
-Unhandled/Unknown Error in engine.process_error
-EXC_TYPE = {type(e)}
-EXC_MSG = {str(e).strip()}
-
-ERROR_CODE = {error_code}
-ERROR_MSG = {error_mesg}
-
-SQL_STMT = {sql_stmt}
-SQL_PARAMS = {sql_params}
-
-{traceback.format_exc()}
-"""
-        print(msg)
+        logger.error(
+            "Unhandled/Unknown Error in engine.process_error",
+            exc_info=True,
+            extra={
+                "error_code": error_code,
+                "error_msg": error_mesg,
+                "sql_stmt": sql_stmt,
+                "sql_params": sql_params,
+            },
+        )
         raise
